package main;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.ArrayList;

import infostuff.AboutDialog;
import infostuff.CmdHelpDialog;
import javafx.application.Platform;
import javafx.geometry.Insets;
import javafx.scene.Node;
import javafx.scene.control.Button;
import javafx.scene.control.CheckBox;
import javafx.scene.control.ComboBox;
import javafx.scene.control.Label;
import javafx.scene.control.Menu;
import javafx.scene.control.MenuBar;
import javafx.scene.control.MenuItem;
import javafx.scene.control.SeparatorMenuItem;
import javafx.scene.control.TextArea;
import javafx.scene.control.Tooltip;
import javafx.scene.input.Dragboard;
import javafx.scene.input.TransferMode;
import javafx.scene.layout.HBox;
import javafx.scene.layout.VBox;
import javafx.scene.text.Font;
import javafx.scene.text.FontWeight;
import javafx.stage.FileChooser;
import javafx.stage.FileChooser.ExtensionFilter;
import javafx.stage.Stage;
import scriptparser.DialogExtractor;
import scriptparser.DialogImporter;
import scriptparser.OffsetRecords;
import scriptparser.ScriptParser;
import scriptparser.ScriptWriter;

public class Menus {
	
	static String version = "v0.30a";

	static Stage winRef;
	static String originalTitle;
	static MenuBar menuBar;

	static Menu fileMenu;
	static MenuItem newMenu;
	static MenuItem closeMnuItm;
	static MenuItem openMnuItm;
	static MenuItem openTxtMnuItm;
	static MenuItem saveMnuItm;
	static MenuItem saveTxtMnuItm;
	static MenuItem exitMnuItm;
	
	static Menu editMenu;
	static MenuItem removeMarkersMnuItm;
	
	static Menu toolsMenu;
	static MenuItem validateMnuItm;
	static MenuItem extractDialogueMnuItm;
	static MenuItem importDialogueMnuItm;
	
	static Menu helpMenu;
	static MenuItem aboutMnuItm;
	static MenuItem cmdHelpMnuItm;

	static TextArea mainText;
	
	static File selectFile;
	static File saveFile;
	
	//left panel stuff
	static VBox leftPanel;
	static CheckBox showOffsetCheck;
	static CheckBox showOpCodeCheck;
	static CheckBox processJumps;
	static ComboBox<String> engineVersion;
	static CheckBox skipLineMarkers;
	

	public static MenuBar getMenus() {

		mainText = new TextArea();
		
		menuBar = new MenuBar();
		fileMenu = new Menu("File");
		newMenu = new MenuItem("New Project");		
		openMnuItm = new MenuItem("Open Script");
		saveMnuItm = new MenuItem("Save As Script");
		saveMnuItm.setDisable(true);
		openTxtMnuItm = new MenuItem("Open Text file");
		saveTxtMnuItm = new MenuItem("Save As Text file");
		saveTxtMnuItm.setDisable(true);
		exitMnuItm = new MenuItem("Exit");
		
		editMenu = new Menu("Edit");
		//editMenu.setDisable(true);
		removeMarkersMnuItm = new MenuItem("Remove Line Markers");
		editMenu.getItems().addAll(removeMarkersMnuItm);
		
		toolsMenu = new Menu("Tools");
		validateMnuItm = new MenuItem("Validate Code");
		extractDialogueMnuItm = new MenuItem("Extract Dialog Only");
		importDialogueMnuItm = new MenuItem("Import Dialog Text File");
		toolsMenu.getItems().addAll(validateMnuItm, new SeparatorMenuItem(),  extractDialogueMnuItm, importDialogueMnuItm);
		validateMnuItm.setDisable(true);
		
		helpMenu = new Menu("Help");
		aboutMnuItm = new MenuItem("About");
		cmdHelpMnuItm = new MenuItem("Commands Help");
		helpMenu.getItems().addAll(aboutMnuItm, cmdHelpMnuItm);
		//helpMenu.getItems().addAll(aboutMnuItm);

		fileMenu.getItems().addAll(newMenu, new SeparatorMenuItem(), openMnuItm, saveMnuItm, new SeparatorMenuItem(), openTxtMnuItm, saveTxtMnuItm, new SeparatorMenuItem(), exitMnuItm);
		menuBar.getMenus().addAll(fileMenu, editMenu, toolsMenu, helpMenu);

		setMenuActions();
		
		return menuBar;
	}

	private static void setMenuActions() {
		
		openMnuItm.setOnAction((e)->{
			FileChooser fileChooser = new FileChooser();
			fileChooser.setTitle("Open AGS Script");
			if (selectFile == null) {
				fileChooser.setInitialDirectory(new File("."));
			} else {
				fileChooser.setInitialDirectory(selectFile.getParentFile());
			}
			fileChooser.getExtensionFilters().addAll(
					new ExtensionFilter("Script Files", "*.scp"),
					new ExtensionFilter("All Files", "*.*"));
			selectFile = fileChooser.showOpenDialog(winRef);
			if (selectFile != null) {
				ScriptParser sp = new ScriptParser(selectFile);
				mainText.setText(sp.getRawData());
				
				if(processJumpsBool()) {
					mainText.setText(OffsetRecords.processJumps(mainText.getText()));
				}
				
				winRef.setTitle(originalTitle + " - " + selectFile.getName());
				saveMnuItm.setDisable(false);
				saveTxtMnuItm.setDisable(false);
				
			}
		});
		
		openTxtMnuItm.setOnAction((e)->{
			FileChooser fileChooser = new FileChooser();
			fileChooser.setTitle("Open Text File");
			if (selectFile == null) {
				fileChooser.setInitialDirectory(new File("."));
			} else {
				fileChooser.setInitialDirectory(selectFile.getParentFile());
			}
			fileChooser.getExtensionFilters().addAll(
					new ExtensionFilter("Text Files", "*.txt"),
					new ExtensionFilter("All Files", "*.*"));
			selectFile = fileChooser.showOpenDialog(winRef);
			if (selectFile != null) {
				try {
					new ScriptParser(selectFile, mainText);
				} catch (IOException e1) {
					e1.printStackTrace();
				}
				
				winRef.setTitle(originalTitle + " - " + selectFile.getName());
				saveMnuItm.setDisable(false);
				saveTxtMnuItm.setDisable(false);
				
			}
		});
		
		//Save Button
		saveMnuItm.setOnAction((e)->{
			FileChooser fileChooser = new FileChooser();
			fileChooser.setTitle("Save AGS Script");
			if (selectFile == null) {
				fileChooser.setInitialDirectory(new File("."));
			} else {
				fileChooser.setInitialDirectory(selectFile.getParentFile());
			}
			fileChooser.getExtensionFilters().addAll(
					new ExtensionFilter("Script Files", "*.scp"),
					new ExtensionFilter("All Files", "*.*"));
			saveFile = fileChooser.showSaveDialog(winRef);
			
			if (selectFile != null) {
				try {
					ScriptWriter sw = new ScriptWriter(saveFile);
					
					ArrayList<String> strArr = new ArrayList<>();
					for (String s : mainText.getText().split("\n")) {
						strArr.add(s);
					}
					
					sw.parseScript(strArr);
					sw.close();
					
					
				} catch (FileNotFoundException e1) {
					e1.printStackTrace();
				}				
			}
		});
		
		saveTxtMnuItm.setOnAction((e)->{
			FileChooser fileChooser = new FileChooser();
			fileChooser.setTitle("Save as Text File");
			if (selectFile == null) {
				fileChooser.setInitialDirectory(new File("."));
			} else {
				fileChooser.setInitialDirectory(selectFile.getParentFile());
			}
			fileChooser.getExtensionFilters().addAll(
					new ExtensionFilter("Text Files", "*.txt"),
					new ExtensionFilter("All Files", "*.*"));
			saveFile = fileChooser.showSaveDialog(winRef);
			
			if (selectFile != null) {
				try {
					ScriptWriter sw = new ScriptWriter(saveFile, mainText.getText());
					sw.close();
					
				} catch (Exception e1) {
					e1.printStackTrace();
				} 			
			}
		});
		
		mainText.setOnDragOver((e)->{
			Dragboard db = e.getDragboard();
			if (db.hasFiles() && db.getFiles().get(0).getName().contains(".scp") || db.getFiles().get(0).getName().contains(".txt")) {
				e.acceptTransferModes(TransferMode.COPY_OR_MOVE);
			}
			e.consume();
		});
		
		mainText.setOnDragDropped((e)-> {
			Dragboard db = e.getDragboard();
			boolean success = false;
			if (db.hasFiles()) {
				success = true;
				File tempFile = db.getFiles().get(0);
				if (tempFile.getName().contains(".scp")) {
					selectFile = db.getFiles().get(0);
					
					ScriptParser sp = new ScriptParser(selectFile);
					mainText.setText(sp.getRawData());
					if(processJumpsBool()) {
						mainText.setText(OffsetRecords.processJumps(mainText.getText()));
					}

					winRef.setTitle(originalTitle + " - " + selectFile.getName());
					saveMnuItm.setDisable(false);
					saveTxtMnuItm.setDisable(false);
				}  else {
					success = false;
				}
			}
			e.setDropCompleted(success);
			e.consume();
			
		});
		
		exitMnuItm.setOnAction((e)->{
			Platform.exit();
		});
		
		aboutMnuItm.setOnAction(e->{
			new AboutDialog().showAndWait();
		});
		
		cmdHelpMnuItm.setOnAction(e->{
			new CmdHelpDialog().show();
		});
		
		extractDialogueMnuItm.setOnAction(e->{
			FileChooser fileChooser = new FileChooser();
			fileChooser.setTitle("Save Dialog to File...");
			
			if (selectFile == null) {
				fileChooser.setInitialDirectory(new File("."));
			} else {
				fileChooser.setInitialFileName(selectFile.getName().substring(0, selectFile.getName().length()-4) +  "_dialog.txt");
				fileChooser.setInitialDirectory(selectFile.getParentFile());
			}
			fileChooser.getExtensionFilters().addAll(
					new ExtensionFilter("Text Files", "*.txt"),
					new ExtensionFilter("All Files", "*.*"));
			saveFile = fileChooser.showSaveDialog(winRef);
			
			if (selectFile != null) {
				try {
					DialogExtractor de = new DialogExtractor(saveFile, mainText.getText());
					de.processText();
					
				} catch (Exception e1) {
					e1.printStackTrace();
				} 			
			}
		});
		
		importDialogueMnuItm.setOnAction(e->{
			FileChooser fileChooser = new FileChooser();
			fileChooser.setTitle("Import Dialog File...");
			
			if (selectFile == null) {
				fileChooser.setInitialDirectory(new File("."));
			} else {
				//fileChooser.setInitialFileName(selectFile.getName().substring(0, selectFile.getName().length()-4) +  "_dialog.txt");
				fileChooser.setInitialDirectory(selectFile.getParentFile());
			}
			
			fileChooser.getExtensionFilters().addAll(
					new ExtensionFilter("Text Files", "*.txt"),
					new ExtensionFilter("All Files", "*.*"));
			
			File openFile = fileChooser.showOpenDialog(winRef);
			
			if (openFile != null) {
				try {
										
					DialogImporter di = new DialogImporter(openFile, mainText.getText());
					di.processText();
					
					//di.replaceText();
					
					mainText.setText(di.replaceText());
					
										
				} catch (Exception e1) {
					e1.printStackTrace();
				} 			
			}
			
		});
		
		removeMarkersMnuItm.setOnAction(e->{
			if (selectFile != null) {
				String temp = mainText.getText().replace("\n\n", "\n");
				ArrayList<String> strArr = new ArrayList<>();
				
				for (String line : temp.split("\n")) {
					if(line.contains("LineMark")) {
						continue;
					}
					if (line.contains("Return")) {
						line = line + "\n";
					}
					strArr.add(line);
				}
				
				mainText.setText(String.join("\n", strArr));
				
				
			} else {
				// add notification here
			}
		});
		
	}

	public static Node getCenter() {
		HBox hBox = new HBox();
		mainText.setFont(Font.font("Lucida Console",FontWeight.EXTRA_BOLD , 22.0));  //font size
		mainText.setStyle("-fx-text-fill: white; \n-fx-control-inner-background: black;");
		mainText.setPrefColumnCount(250);
		hBox.getChildren().addAll( mainText);
		return hBox;
	}
	
	public static VBox getLeft() {
		leftPanel = new VBox();
		leftPanel.setPrefWidth(140);
		leftPanel.setStyle("-fx-background-color: #333;");
		
		HBox addButtonPanel = new HBox();
		Button addCodeButton = new Button("Add Code");
		
		HBox.setMargin(addCodeButton, new Insets(10));
		
		//options panel stuff
		showOffsetCheck = new CheckBox("Show Offsets");
		showOpCodeCheck = new CheckBox("Show OpCodes");
		processJumps = new CheckBox("Process Jumps");
		skipLineMarkers = new CheckBox("Skip LineMarkers");
		showOffsetCheck.setStyle("-fx-text-fill: white;");
		showOpCodeCheck.setStyle("-fx-text-fill: white;");
		processJumps.setStyle("-fx-text-fill: white;");
		skipLineMarkers.setStyle("-fx-text-fill: white;");
		
		engineVersion = new ComboBox<>();
		engineVersion.setStyle("-fx-text-fill: white; \n-fx-control-inner-background: black;");
		
		setUpGameEnginesComboBox();
		setUpComboActions();
		
		
		processJumps.setTooltip(new Tooltip("Generate jump tokens.\nYou'll be able to edit the script\nwithout breaking the jumps offsets."));
		skipLineMarkers.setTooltip(new Tooltip("Removes Line Markers, seems okay to remove."));
		
		
		HBox showOffsetsGroup = new HBox(showOffsetCheck);
		HBox showOpCodesGroup = new HBox(showOpCodeCheck);
		HBox processJumpsGroup = new HBox(processJumps);
		HBox skipLineMarkersGroup = new HBox(skipLineMarkers);
		
		Label engineVersionLabel = new Label("Engine Version:");
		engineVersionLabel.setStyle("-fx-text-fill: white;");
		VBox engineVersionGroup = new VBox(engineVersionLabel, engineVersion);
		
		
		VBox optionsPanel = new VBox(10, showOffsetsGroup, showOpCodesGroup, 
				processJumpsGroup, engineVersionGroup, skipLineMarkersGroup);
		VBox.setMargin(optionsPanel, new Insets(10));
		
		
		//Set items checked 
		showOffsetCheck.setSelected(true);
		processJumps.setSelected(true);
		
		addButtonPanel.getChildren().add(addCodeButton);
		leftPanel.getChildren().addAll(addButtonPanel, optionsPanel);
		
		//add Add Button menu
		
		//addButtonPanel.setOnContextMenuRequested(arg0);
		
		
		return leftPanel;
	}
	
	private static void setUpGameEnginesComboBox() {
		ArrayList<String> engines = new ArrayList<>();
		engines.add("1.20030523 (VIPER -M3- 3.2)");
		engines.add("2.20040326 (Love Fetish ~RL~)");
		engines.add("3.20040514 (Love Fetish `pCYҁ`)");
		engines.add("4.20040528 (Love Fetish ``Iij[ҁ`)");
		engines.add("5.20040625 (Love Fetish `rtF`RLҁ`)");
		engines.add("6+ AGS");
		engines.add("6+ Encrypted Script (ver.1)");
		engines.add("6+ Encrypted Script (ver.2)");
		
		engineVersion.getItems().addAll(engines);
		engineVersion.getSelectionModel().selectFirst();
	}

	private static void setUpComboActions() {
		engineVersion.setOnAction((E)->{
			
		});
	}

	public static boolean getOffsetsBool() {
		return showOffsetCheck.isSelected();
	}
	
	public static boolean getOpCodesBool() {
		return showOpCodeCheck.isSelected();
	}
	
	public static boolean processJumpsBool() {
		return processJumps.isSelected();
	}
	
	public static boolean getLineMarkersBool() {
		return skipLineMarkers.isSelected();
	}
	
	public static int getEngineVersionValue() {
		return engineVersion.getSelectionModel().getSelectedIndex()+1;
	}
	
	public static void setStage(Stage window) {
		winRef = window;
		originalTitle = window.getTitle() + " " + version;
		window.setTitle(originalTitle);
	}

	public static boolean getLineNumbersBool() {
		return false;
	}

	public static String getVersion() {
		return version;
	}

}
