package scriptparser;

import java.util.ArrayList;
import java.util.TreeMap;

import main.Menus;

public class OffsetRecords {
	private static TreeMap<Integer, Integer> offsetRecords;
	private static ArrayList<Integer> jumpTargets;
	
	private static TreeMap<Integer, Integer> linesWithOffsets;
	private static TreeMap<Integer, String> labeledJumps;
	
	
	private static int jumpNum;
	
	private  OffsetRecords() {
	}
	
	public static void setOffsetRecords(TreeMap<Integer, Integer> source, TreeMap<Integer, Integer> linesOffsets) {
		offsetRecords = source;
		linesWithOffsets = linesOffsets;
		labeledJumps = new TreeMap<>();
		
		generateLabels();
	}
	
	public static void setOffsetRecords(ArrayList<Integer> source, TreeMap<Integer, Integer> linesOffsets) {
		jumpTargets = source;
		linesWithOffsets = linesOffsets;
		labeledJumps = new TreeMap<>();
		
		generateLabelsWithArrayList();
	}
	
	
	public static void generateLabelsWithArrayList() {
		
		String prefix = "JUMP";
		jumpNum = 1;
		
		jumpTargets.forEach(key->{
			//String for jump labels generated here
			//System.out.println("key: " + key);
			labeledJumps.put(key, prefix + get4Digits(jumpNum++));
		});
		
	}
	
	public static void generateLabels() {
		
		String prefix = "JUMP";
		jumpNum = 1;
		offsetRecords.keySet().iterator().forEachRemaining(key ->{
			//String for jump labels generated here
			//System.out.println("key: " + key);
			
			labeledJumps.put(key, prefix + get4Digits(jumpNum++));			
		});
	}
	
	private static String get4Digits(int x) {
		String str = Integer.toString(x);
		while (str.length() < 4) {
			str = "0" + str;
		}
		return str;
	}
	
	public static String processJumps(String scriptText) {
		ArrayList<String> lines = new ArrayList<>();
		int lineNo = 1;
		
		String[] scriptArray = scriptText.split("\n");
		
		lineNo = 1;
		
		for (String s : scriptArray) { 
			if(Menus.getLineNumbersBool()) {
				lines.add("|" + lineNo++ +  s);
			} else {
				lines.add(s);
			}
			
		}
		
		//System.out.println("lines found: " + lines.size());
		
		int linesTotal = lines.size();
		
		String workingLine;
		
		for (int x = linesTotal; x > 0; x-- ) {  //add in JumpLabels
			int value = linesWithOffsets.get(x);
			//System.out.println("Value checked: " + value + " -> " + Integer.toHexString(value));
			if( labeledJumps.containsKey(value)) {
				//Jump labels inserted here 
				//System.out.println("Jump Added at : " + value + " -> " + Integer.toHexString(value));
				
				/*
				System.out.println("x value:      " + x );
				System.out.println("Line Text     " + lines.get(x-1) );
				System.out.println("Line Offsets  " + Integer.toHexString(linesWithOffsets.get(x)) );
				System.out.println("Labeled Jump: " + labeledJumps.get(linesWithOffsets.get(x)) + "\n" );
				*/
				
				lines.add(x-1, "@" + labeledJumps.get(linesWithOffsets.get(x)));
			}
		}
		
		linesTotal = lines.size()-1;
		
		for (int x = linesTotal; x >= 0; x-- ) {  // replace Jump lines
			if (lines.get(x).contains("ConditionalJump")) {
				workingLine = lines.get(x);
				
				String offset = workingLine.substring(workingLine.lastIndexOf(",")+3, workingLine.lastIndexOf(")"));
				int offsetInt = Integer.parseUnsignedInt(offset, 16);
				String fixedLine = workingLine.substring(0, workingLine.lastIndexOf(",")+1);
				String fixedLineEnd = workingLine.substring(workingLine.lastIndexOf(")"));
				
				String finalString = fixedLine + labeledJumps.get(offsetInt) + fixedLineEnd;
				
				lines.set(x, finalString);
				
			}
			
			if (lines.get(x).contains("GoTo")) {
				//System.out.println("Line: #" + x + " - " + lines.get(x));
				workingLine = lines.get(x);
				
				String offset = workingLine.substring(workingLine.lastIndexOf("x")+1, workingLine.lastIndexOf(")")).trim();
				int offsetInt = Integer.parseUnsignedInt(offset, 16);
				String fixedLine = workingLine.substring(0, workingLine.lastIndexOf("(")+1);
				String fixedLineEnd = workingLine.substring(workingLine.lastIndexOf(")"));
				String finalString = fixedLine + labeledJumps.get(offsetInt) + fixedLineEnd;
				lines.set(x, finalString);
				
			}
			
			if (lines.get(x).contains("JumpTo")) {
				//System.out.println("Line: #" + x + " - " + lines.get(x));
				workingLine = lines.get(x);
				
				String offset = workingLine.substring(workingLine.lastIndexOf("x")+1, workingLine.lastIndexOf(")")).trim();
				int offsetInt = Integer.parseUnsignedInt(offset, 16);
				String fixedLine = workingLine.substring(0, workingLine.lastIndexOf("(")+1);
				String fixedLineEnd = workingLine.substring(workingLine.lastIndexOf(")"));
				String finalString = fixedLine + labeledJumps.get(offsetInt) + fixedLineEnd;
				lines.set(x, finalString);
			}
			
		}
		
		return String.join("\n", lines).replace("Return;\n", "Return;\n\n");
	}
	
}
