package scriptparser;

import java.io.BufferedInputStream;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.UnsupportedEncodingException;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.util.ArrayList;
import java.util.Collections;
import java.util.TreeMap;

import javafx.scene.control.TextArea;
import main.Menus;
import scriptparser.ErrorTracker.ErrorType;

public class ScriptParser {
	
	byte[] data;
	
	ByteBuffer bb;
	StringBuilder sb;
	BufferedInputStream bis;
	
	String offset = "";
	int pos = 0;
	String opCode = "";
	
	TreeMap<Integer, Integer> offsetsRecord;
	ArrayList<Integer> offsetRecords;
	
	TreeMap<Integer, Integer> offsetTracker;
	
	
	int lineNum = 1;
	
	
	
	public ScriptParser(File file) { //reading in script file
		
		ErrorTracker.prepErrorTracking();   
		
		try {
			bis = new BufferedInputStream(new FileInputStream(file));
			data = new byte[bis.available()];
			bis.read(data);
			bis.close();
			bis = null;
			
			bb = ByteBuffer.wrap(data);
			bb.order(ByteOrder.nativeOrder());
			
			getMapReady();
			performDescrypt();
			
		} catch (IOException e) {
			e.printStackTrace();
		}
	}

	private void performDescrypt() {
		if (Menus.getEngineVersionValue() >= 7) {
			AGS_Decrypt.resetDecrypt();
			
			byte[] decryptedData = new byte[bb.remaining()];
			int x = 0;
			
			
			while (bb.hasRemaining()) {
				decryptedData[x++] = (byte) AGS_Decrypt.decryptInt(bb.get());
				
				//System.out.print(get2Bytes(decryptedData[x-1]&0xFF) + " ");
				if (x % 16 == 0) {
					//System.out.print("\n");
				}
			}
			bb = ByteBuffer.wrap(decryptedData);
			bb.order(ByteOrder.nativeOrder());
			bb.position(0);
			
		}
	}

	public ScriptParser(File selectFile, TextArea mainText) throws IOException {  //reading in a text file
		InputStreamReader isr = new InputStreamReader(new FileInputStream(selectFile), "UTF-8");
		BufferedReader br = new BufferedReader(isr);
		getMapReady();
		
		String sRead = null;
		String buffer = "";
		
		do {
			sRead = br.readLine();
			if (sRead == null) break;
			buffer += sRead + "\n";
		} while (sRead != null);
		
		mainText.setText(buffer);
		
		try {
			br.close();
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

	}

	public String getRawData() {
		sb = new StringBuilder();
		while(bb.hasRemaining()) {
			pos = bb.position();
			//collecting lines and offset
			offsetTracker.put(lineNum++, pos);
			
			int x = bb.get() & 0xFF;
			
			offset = "";
			opCode = "";
			
			if (Menus.getOffsetsBool()) {
				offset = "|" + get2Bytes(pos) + "| ";
			} 

			if (Menus.getOpCodesBool()) {
				opCode = getByte(x) + " | ";
			}

			String command = offset + opCode + getCommand(x);

			if (!processCommand(x)) {
				sb.append(command + "\n");
				ErrorTracker.addError(ErrorType.PARSING, command + "\n");
			}
		}
		
		setOffsetRecords();
		//printOffsetsRecord3();
		
		if (ErrorTracker.checkErrors()) {
			ErrorTracker.showErrorDialog();
		}
		
		return sb.toString();
		
	}

	private boolean processCommand(int x) {
		String cmd = "";
		String dialogue = "";
		int a, b, c, d;
		
		int debugOffset = bb.position()-1;
		
		byte bt = 0;
		byte[] rawBytes = null;
		int bytePos = 0;
		
		
		
		switch(x) {
		case 0x00:
			if (Menus.getEngineVersionValue() > 2 && bb.remaining() > 1) {
				cmd = "LineMark ( ";
				a = bb.getShort();
				
				cmd += a + " );\n";
			} else {
				cmd = "cmd_nop ();\n";
			}
			
			if (!Menus.getLineMarkersBool()) {
				sb.append(offset + opCode + cmd);	
			}
			
			break;
		
		case 0x01:
			cmd = "DisplayText (\"";
			dialogue = "";
			bt = 0;
			
			rawBytes = new byte[80];
			bytePos = 0;
			
			do {
				bt = bb.get();
				rawBytes[bytePos++] = bt;
			} while (bt != 0x00 && bytePos < rawBytes.length);
			
			
			try {
				// TODO: this gets rid of the Whitespace at the beginning, Fix this.
				if (new String(rawBytes, "Shift-JIS").indexOf(0x00) != -1) {
					dialogue = new String(rawBytes, "Shift-JIS").substring(0, new String(rawBytes, "Shift-JIS").indexOf(0x00));
				} else {
					dialogue = new String(rawBytes, "Shift-JIS");
				}
				
				
				//dialogue = new String(rawBytes, "Shift-JIS").trim();   
				
			} catch (UnsupportedEncodingException e) {
				e.printStackTrace();
			}
			
			cmd += dialogue + "\");\n";
			
			sb.append(offset + opCode + cmd);
			break;
			
		case 0x02:
			cmd = "FlushBuffers ();\n";
			sb.append(offset + opCode + cmd);
			break;
			
		case 0x03:
			cmd = "cmd_0x03 ();\n";
			sb.append(offset + opCode + cmd);
			break;
		
		case 0x04:
			cmd = "SetTextBox (";
			a = bb.getShort();
			b = bb.getShort();
			c = bb.getShort();
			d = bb.getShort();
			cmd += a + "," + b + "," + c + ","+ d + ");\n";
			sb.append(offset + opCode + cmd);
			break;
			
		case 0x05:
			cmd = "SetCursorPosition (";
			a = bb.getShort();
			b = bb.getShort();
			cmd += a + "," + b + ");\n";
			sb.append(offset + opCode + cmd);
			break;
			
		case 0x07:
			if (Menus.getEngineVersionValue() > 4) {
				cmd = "DisplayText_f (";
				a = bb.getShort() & 0xFFFF;
				
				dialogue = "";
				bt = 0;
				
				rawBytes = new byte[80];
				bytePos = 0;
				
				do {
					bt = bb.get();
					rawBytes[bytePos++] = bt;
					
					
				} while (bt != 0x00 && bytePos < rawBytes.length);

				try {
					dialogue = new String(rawBytes, "Shift-JIS").substring(0, new String(rawBytes, "Shift-JIS").indexOf(0x00));
					//dialogue = new String(rawBytes, "Shift-JIS").trim();
				} catch (UnsupportedEncodingException e) {
					e.printStackTrace();
				}
				
				cmd += get2Bytes(a) + ", \""  + dialogue + "\");\n";
				
				sb.append(offset + opCode + cmd);
				break;
			}
			return false;
			
			
		case 0x0A:
			cmd = "ExecuteScript (";
			rawBytes = new byte[0x10];
			bytePos = 0;
			
			do {
				bt = bb.get();
				rawBytes[bytePos++] = bt;
			} while (bt != 0x00 && bytePos < rawBytes.length);
			
			dialogue = new String(rawBytes).trim();
			cmd += dialogue + ".scp);\n";
			sb.append(offset + opCode + cmd);
			break;
			
		case 0x0B:
			cmd = "GetSystemTime();\n";
			sb.append(offset + opCode + cmd);
			break;
			
		case 0x0C:
			cmd = "MouseOverEvent (";
			a = bb.getShort() & 0xFFFF;
			b = bb.getShort() & 0xFFFF;
			c = bb.getShort() & 0xFFFF;
			d = bb.getShort() & 0xFFFF;
			
			cmd += a + ", " + b + ", " + c + ", " + d + " );\n";
			
			sb.append(offset + opCode + cmd);
			break;
			
		case 0x0F:
			cmd = "EnableSaveMenu (";
			a = bb.get();
			cmd += a +");\n";
			sb.append(offset + opCode + cmd);
			break;
			
		case 0x13:
			cmd = "cmd_0x13 ( ";
			a = bb.getShort() & 0xFFFF;
			cmd += get2Bytes(a) + " );\n";
			
			sb.append(offset + opCode + cmd);
			break;
			
		case 0x14:
			cmd = "GoTo ( ";
			a = bb.getShort() & 0xFFFF;
			cmd += get2Bytes(a) +" );\n";
			sb.append(offset + opCode + cmd);
			
			addJumpCode(a, debugOffset);
			//offsetsRecord.put(a, a);
			break;
			
		case 0x15:
			cmd = "JumpTo ( ";
			a = bb.getShort() & 0xFFFF;
			cmd += get2Bytes(a) +" );\n";
			sb.append(offset + opCode + cmd);
			addJumpCode(a, debugOffset);
			//offsetsRecord.put(a, a);
			break;
			
		case 0x16:
			cmd = "Return;\n";
			sb.append(offset + opCode + cmd);
			break;
			
		case 0x17:
			cmd = "cmd_0x17 ( ";
			a = bb.get() & 0xFF;
			b = bb.get() & 0xFF;
			
			cmd += getByte(a) + ", " + getByte(b) +" );\n";
			sb.append(offset + opCode + cmd);
			break;
			
		case 0x18:
			cmd = "cmd_0x18 ( ";
			a = bb.get() & 0xFF;
			b = bb.get() & 0xFF;
			c = bb.get() & 0xFF;
			
			cmd+= a + "," + b + "," + c + " );\n"; 
			sb.append(offset + opCode + cmd);
			break;
			
		case 0x19:
			cmd = "ConditionalJump (";
			a = bb.get() & 0xFF;
			b = bb.get() & 0xFF;
			c = bb.get() & 0xFF;
			d = bb.getShort() & 0xFFFF;
			
			cmd+= a + "," + b + "," + c + "," + get2Bytes(d) + ");\n"; 
			sb.append(offset + opCode + cmd);
			addJumpCode(d, debugOffset);
			break;
			
		case 0x1A:
			if(Menus.getEngineVersionValue() >= 7) {
				cmd = "cmd_0x1A (";
				a = bb.get() & 0xFF;
				b = bb.getShort() & 0xFFFF;
				c = bb.getShort() & 0xFFFF;
				
				cmd+= getByte(a) + "," + get2Bytes(b) + "," + get2Bytes(c) + ");\n"; 
				
				sb.append(offset + opCode + cmd);
				break;
			}
			
			return false;
			
		case 0x1E:
			cmd = "LoadANI ( ";
			rawBytes = new byte[512];
			bytePos = 0;
			
			do {
				bt = bb.get();
				rawBytes[bytePos++] = bt;
			} while (bt != 0x00 && bytePos < rawBytes.length);
			
			dialogue = new String(rawBytes).trim();
			cmd += dialogue + ".ani );\n";
			sb.append(offset + opCode + cmd);
			break;
			
		case 0x1F:
			cmd = "TweenFrames ( ";
			a = bb.get();
			b = bb.getShort();
			c = bb.getShort();
			
			cmd += a + "," + b + "," + c + " );\n";
			sb.append(offset + opCode + cmd);
			break;
			
		case 0x20:
			cmd = "cmd_0x20 ( ";
			a = bb.getShort() & 0xFFFF;
			
			cmd += a + " );\n";
			
			sb.append(offset + opCode + cmd); 
			break;
			
		case 0x21:
			cmd = "WaitForFrame ( ";
			a = bb.get() & 0xFF;
			b = bb.getShort() & 0xFFFF;
			
			cmd += a + " , " + b + " );\n";
			sb.append(offset + opCode + cmd);
			break;
			
		case 0x22:
			cmd = "cmd_0x22 ( ";
			a = bb.getShort() & 0xFFFF;
			
			cmd += a + " );\n";
			sb.append(offset + opCode + cmd);
			break;	
			
		case 0x23:
			cmd = "cmd_0x23 ( ";
			a = bb.getShort() & 0xFFFF;
			
			cmd += get2Bytes(a) + " );\n";
			
			sb.append(offset + opCode + cmd);
			break;
			
		case 0x24: // unknown command
			cmd = "cmd_0x24 ( ";
			a = bb.get() & 0xFF;
						
			cmd += getByte(a) + " );\n";
			sb.append(offset + opCode + cmd);
			break;
			
		case 0x25:
			if (Menus.getEngineVersionValue() > 2) {
				cmd = "cmd_0x25 ();\n";
				sb.append(offset + opCode + cmd);
				break;
			}
			return false;
			
		case 0x27:
			if (Menus.getEngineVersionValue() > 2) {
				cmd = "LoadANI_f ( ";
				a = bb.get() & 0xFF;
				
				rawBytes = new byte[128];
				bytePos = 0;
				
				do {
					bt = bb.get();
					rawBytes[bytePos++] = bt;
				} while (bt != 0x00 && bytePos < rawBytes.length);
				
				dialogue = new String(rawBytes).trim();
				
				cmd += a + ", " + dialogue + ".ani );\n";
				sb.append(offset + opCode + cmd);
				
				break;
			}
			
			return false;
			
		case 0x28:
			cmd = "FadeFromBlack ( ";
			a = bb.get() & 0xFF;
			
			cmd += a + " );\n";
			sb.append(offset + opCode + cmd);
			break;
			
		case 0x29:
			cmd = "FadeToBlack ( ";
			a = bb.get() & 0xFF;
			
			cmd += a + " );\n";
			sb.append(offset + opCode + cmd);
			break;
			
		case 0x2A:
			cmd = "FadeFromWhite ( ";
			a = bb.get() & 0xFF;
			
			cmd += a + " );\n";
			sb.append(offset + opCode + cmd);
			break;
			
		case 0x2B:
			cmd = "FadeToWhite ( ";
			a = bb.get() & 0xFF;
			
			cmd += a + " );\n";
			sb.append(offset + opCode + cmd);
			break;
			
		case 0x2C:
			cmd = "cmd_0x2C ( ";
			a = bb.get() & 0xFF;
			
			cmd += getByte(a) + " );\n";
			sb.append(offset + opCode + cmd);
			break;
			
		case 0x2D: // unknown command
			cmd = "cmd_0x2D ();\n";
			sb.append(offset + opCode + cmd);
			break;
			
		case 0x2E:
			cmd = "cmd_0x2E ();\n";
			sb.append(offset + opCode + cmd);
			break;
			
		case 0x2F:
			cmd = "cmd_0x2F ( ";
			a = bb.get() & 0xFF;
			
			
			cmd += getByte(a) + " );\n";
			sb.append(offset + opCode + cmd);
			break;
			
		case 0x30:
			cmd = "cmd_0x30 ( ";
			a = bb.get() & 0xFF;
			
			
			cmd += getByte(a) + " );\n";
			sb.append(offset + opCode + cmd);
			break;
			
		case 0x31: // unknown command
			cmd = "cmd_0x31 ( ";
			a = bb.get() & 0xFF;
						
			cmd += a + " );\n";
			sb.append(offset + opCode + cmd);
			break;
		
		case 0x32:
			cmd = "SetRenderSize (";
			a = bb.getShort();
			b = bb.getShort();
			c = bb.getShort();
			d = bb.getShort();
			cmd += a + "," + b + "," + c + ","+ d + ");\n";
			sb.append(offset + opCode + cmd);
			break;
			
		case 0x33:
			cmd = "DrawImage ( ";
			a = bb.getShort();
			b = bb.getShort();
			c = bb.getShort();
			cmd += a + "," + b + "," + c + " );\n";
			sb.append(offset + opCode + cmd);
			break;
			
		case 0x34:
			cmd = "LoadCG ( ";
			
			rawBytes = new byte[0x20];
			bytePos = 0;
			
			do {
				bt = bb.get();
				rawBytes[bytePos++] = bt;
			} while (bt != 0x00 && bytePos < rawBytes.length);
			dialogue = new String(rawBytes).trim();
			
			cmd += dialogue + ".cg );\n";
			
			sb.append(offset + opCode + cmd);
			break;
			
		case 0x35:
			cmd = "cmd_0x35 ();\n";
			sb.append(offset + opCode + cmd);
			break;
			
		case 0x36:
			cmd = "cmd_0x36 ();\n";
			sb.append(offset + opCode + cmd);
			break;
			
		case 0x37:
			cmd = "LoadCG_2 ( ";
			
			rawBytes = new byte[0x20];
			bytePos = 0;
			
			do {
				bt = bb.get();
				rawBytes[bytePos++] = bt;
			} while (bt != 0x00 && bytePos < rawBytes.length);
			dialogue = new String(rawBytes).trim();
			
			cmd += dialogue + ".cg );\n";
			
			sb.append(offset + opCode + cmd);
			break;
			
		case 0x39:
			if (Menus.getEngineVersionValue() > 3) {
				cmd = "DrawImage_f ( ";
				a = bb.get() & 0xFF;
				b = bb.getShort() & 0xFFFF;
				c = bb.getShort() & 0xFFFF;
				d = bb.getShort() & 0xFFFF;
				cmd += getByte(a) + "," + get2Bytes(b) + "," + get2Bytes(c) + "," + get2Bytes(d) + " ); unverified code\n";
				sb.append(offset + opCode + cmd);
				
				break;
			}
			return false;
			
		case 0x46:
			cmd = "LoadAudioClip_l ( ";
			a = bb.get();
			rawBytes = new byte[0x20];
			bytePos = 0;
			
			do {
				bt = bb.get();
				rawBytes[bytePos++] = bt;
			} while (bt != 0x00 && bytePos < rawBytes.length);
			dialogue = new String(rawBytes).trim();
			
			cmd += a + " , " +  dialogue + ".pcm );\n";
			
			sb.append(offset + opCode + cmd);
			break;
			
		case 0x47:
			cmd = "cmd_0x47 ();\n";
			sb.append(offset + opCode + cmd);
			break;
			
		case 0x48:
			cmd = "PlaySFX_l ( ";
			a = bb.get() & 0xFF;
			
			rawBytes = new byte[0x20];
			bytePos = 0;
			
			do {
				bt = bb.get();
				rawBytes[bytePos++] = bt;
			} while (bt != 0x00 && bytePos < rawBytes.length);
			dialogue = new String(rawBytes).trim();
			
			cmd += a + " , " +  dialogue + ".pcm );\n";
			
			sb.append(offset + opCode + cmd);
			break;
			
		case 0x49:
			cmd = "cmd_0x49 ();\n";
			sb.append(offset + opCode + cmd);
			break;
			
		case 0x4A:
			cmd = "LoadBGM_2 ( ";
			a = bb.get() & 0xFF;
			rawBytes = new byte[0x20];
			bytePos = 0;
			
			do {
				bt = bb.get();
				rawBytes[bytePos++] = bt;
			} while (bt != 0x00 && bytePos < rawBytes.length);
			dialogue = new String(rawBytes).trim();
			cmd += a + " , " +  dialogue + ".pcm );\n";
			
			sb.append(offset + opCode + cmd);
			
			break;
			
		case 0x4B:
			cmd = "cmd_0x4B ();\n";
			sb.append(offset + opCode + cmd);
			break;
			
		case 0x4C:
			cmd = "cmd_0x4C ( ";
			a = bb.get() & 0xFF;
			cmd += getByte(a) + " );\n";
			
			sb.append(offset + opCode + cmd);
			break;
			
		case 0x4D:
			cmd = "cmd_0x4D ( ";
			a = bb.get() & 0xFF;
			cmd += getByte(a) + " );\n";
			
			sb.append(offset + opCode + cmd);
			break;
			
		case 0x4E:
			cmd = "cmd_0x4E ( ";
			a = bb.get() & 0xFF;
			cmd += getByte(a) + " );\n";
			
			sb.append(offset + opCode + cmd);
			
			break;
			
		case 0x4F:
			if (Menus.getEngineVersionValue() > 2) {
				cmd = "LoadAudioClip_f ( ";
				a = bb.get() & 0xFF;
				b = bb.get() & 0xFF;
				
				rawBytes = new byte[0x20];
				bytePos = 0;
				
				do {
					bt = bb.get();
					rawBytes[bytePos++] = bt;
				} while (bt != 0x00 && bytePos < rawBytes.length);
				dialogue = new String(rawBytes).trim();
				
				cmd += a + ", " + b + " , " +  dialogue + ".pcm );\n";
				
				sb.append(offset + opCode + cmd);
				break;
			}
			return false;
			
		case 0x51:
			if (Menus.getEngineVersionValue() < 3) {
				cmd = "cmd_0x51 ( ";
				a = bb.getShort() & 0xFFFF;
				b = bb.getShort() & 0xFFFF;
				
				cmd += get2Bytes(a) + ", " + get2Bytes(b) + " );\n";
			} else {
				cmd = "cmd_0x51 ( ";
				
				a = bb.get() & 0xFF;
				b = bb.getShort() & 0xFFFF;
				c = bb.getShort() & 0xFFFF;
				
				cmd += getByte(a) + ", " + get2Bytes(b) + ", " + get2Bytes(c) + " );\n";
			}
			
			sb.append(offset + opCode + cmd);
			break;
			
		case 0x52:
			cmd = "cmd_0x52 ();\n";
			sb.append(offset + opCode + cmd);
			break;
			
		case 0x54:  // TODO: verify word arg
			cmd = "cmd_0x54 ( ";
			a = bb.getShort() & 0xFFFF;
			
			cmd += get2Bytes(a) + " );\n";
			sb.append(offset + opCode + cmd);
			break;
			
		case 0x55:
			cmd = "cmd_0x55 ();\n";
			sb.append(offset + opCode + cmd);
			break;
			
		case 0x56:
			cmd = "cmd_0x56 ( ";
			a = bb.getShort() & 0xFFFF;
			
			cmd += get2Bytes(a) + " );\n";
			
			sb.append(offset + opCode + cmd);
			break;
			
		case 0x57:
			cmd = "cmd_0x57 ();\n";
			sb.append(offset + opCode + cmd);
			break;
			
		case 0x58:
			cmd = "cmd_0x58 ( ";
			a = bb.getShort() & 0xFFFF;
			
			cmd += get2Bytes(a) + " );\n";
			sb.append(offset + opCode + cmd);
			break;
			
			
		case 0x59:
			cmd = "cmd_0x59 ( ";
			a = bb.get() & 0xFF;
			b = bb.get() & 0xFF;
			c = bb.get() & 0xFF;
			
			cmd += getByte(a) + ", " + getByte(b) + ", " + getByte(c) + " );\n";
			sb.append(offset + opCode + cmd);
			break;
			
		case 0x5B:
			cmd = "SetFontSize (";
			a = bb.get();
			b = bb.get();
			c = bb.get();
			cmd += a + "," + b + "," + c + ");\n";
			sb.append(offset + opCode + cmd);
			break;
		
		case 0x64:
			cmd = "SetWindowSize (";
			a = bb.getShort();
			b = bb.getShort();
			cmd += a + "," + b + ");\n";
			sb.append(offset + opCode + cmd);
			break;
			
		case 0x68:
			cmd = "SetRenderDimensions (";
			a = bb.getShort();
			b = bb.getShort();
			cmd += a + "," + b + ");\n";
			sb.append(offset + opCode + cmd);
			break;
			
		case 0x69:
			cmd = "LoadSaveFile (";
			a = bb.get() & 0xFF;
			b = bb.get() & 0xFF;
			cmd += a + "," + b + ");\n";
			sb.append(offset + opCode + cmd);
			break;
			
		case 0x6A:
			if(Menus.getEngineVersionValue() >= 7) {
				cmd = "cmd_0x6A (";
				a = bb.get() & 0xFF;
				cmd += a + ");\n";
				
				sb.append(offset + opCode + cmd);
				break;
			}
			return false;
			
		case 0x6B:
			cmd = "StoreText (";
			a = bb.get() & 0xFF;
			
			dialogue = "";
			bt = 0;

			rawBytes = new byte[0x30];
			bytePos = 0;

			do {
				bt = bb.get();
				rawBytes[bytePos++] = bt;
			} while (bt != 0x00 && bytePos < rawBytes.length);

			try {
				dialogue = new String(rawBytes, "Shift-JIS").trim();
			} catch (UnsupportedEncodingException e) {
				e.printStackTrace();
			}

			cmd += a + ",\"" + dialogue + "\");\n";
			sb.append(offset + opCode + cmd);

			break;
			
		case 0x6E:
			cmd = "TransformImageTo ( ";
			a = bb.getShort();
			b = bb.getShort();
			c = bb.getShort();
			d = bb.getShort();
			
			cmd += a + "ms ," + b + "," + c + ","+ d + " );\n";
			sb.append(offset + opCode + cmd);
			break;
			
		case 0x6F:
			cmd = "EnableImagePanRange ( ";
			a = bb.getShort();
			b = bb.getShort();
			c = bb.getShort();
			d = bb.getShort();
			
			cmd += a + "," + b + "," + c + ","+ d + " );\n";
			sb.append(offset + opCode + cmd);
			break;
			
		case 0x70:
			cmd = "WaitForTransform ();\n";
			sb.append(offset + opCode + cmd);
			break;
			
		case 0x71:
			cmd = "cmd_0x71 ();\n";
			sb.append(offset + opCode + cmd);
			break;
			
		case 0x73:
			cmd = "cmd_0x73 ( ";
			a = bb.getShort() & 0xFFFF;
			b = bb.getShort() & 0xFFFF;
			
			cmd += get2Bytes(a) + ", " + get2Bytes(b) + " );\n";
			sb.append(offset + opCode + cmd);
			break;
			
		case 0x78:
			cmd = "cmd_0x78 ( ";
			a = bb.get() & 0xFF;
						
			cmd += a + " );\n";
			sb.append(offset + opCode + cmd);
			break;
			
		case 0x7A:
			cmd = "LoadBGM ( ";
			a = bb.getShort() & 0xFFFF;
			rawBytes = new byte[0x20];
			bytePos = 0;
			
			do {
				bt = bb.get();
				rawBytes[bytePos++] = bt;
			} while (bt != 0x00 && bytePos < rawBytes.length);
			
			dialogue = new String(rawBytes).trim();
			
			cmd += a + " , " + dialogue + ".pcm );\n";
			sb.append(offset + opCode + cmd);
			break;
			
		case 0x7B:
			cmd = "cmd_0x7B ();\n";
			sb.append(offset + opCode + cmd);
			break;
			
		case 0x82:
			if (Menus.getEngineVersionValue() > 2) {
				cmd = "cmd_0x82 ( ";
				a = bb.getShort() & 0xFFFF;
				b = bb.getShort() & 0xFFFF;
				c = bb.getShort() & 0xFFFF;
				d = bb.getShort() & 0xFFFF;
				
				cmd += a + "," + b + "," + c + ","+ d + " );\n";
				sb.append(offset + opCode + cmd);
				break;
			}
			return false;
			
		case 0x83:
			if (Menus.getEngineVersionValue() > 3) {
				cmd = "cmd_0x83 ( ";
				a = bb.get() & 0xFF;
				
				cmd += a + " );\n";
				sb.append(offset + opCode + cmd);
				break;
			}
			return false;
			
		case 0x84:
			if (Menus.getEngineVersionValue() > 2) {
				cmd = "cmd_0x84 ( ";
				a = bb.get() & 0xFF;
				
				cmd += a + " );\n";
				sb.append(offset + opCode + cmd);
				break;
			}
			return false;
			
		case 0x8C:
			if (Menus.getEngineVersionValue() > 2) {
				cmd = "cmd_0x8C ( ";
				a = bb.get() & 0xFF;
				
				cmd += a + " );\n";
				sb.append(offset + opCode + cmd);
				break;
			}
			return false;
			
		case 0x8D:
			if (Menus.getEngineVersionValue() > 2) {
				cmd = "cmd_0x8D ();\n";
				sb.append(offset + opCode + cmd);
				break;
			}
			return false;
			
		case 0x96:
			if (Menus.getEngineVersionValue() > 4) {
				cmd = "cmd_0x96 ( ";
				a = bb.get() & 0xFF;
				b = bb.get() & 0xFF;
				c = bb.get() & 0xFF;
				d = bb.get() & 0xFF;
				cmd += getByte(a) + "," + getByte(b) + "," + getByte(c) + "," + getByte(d) + " ); \t unknown code\n";
				sb.append(offset + opCode + cmd);
				
				break;
			}
			return false;
			
		case 0x99:
			if (Menus.getEngineVersionValue() > 5) {
				cmd = "cmd_0x99 ( ";
				a = bb.get() & 0xFF;
				
				cmd += getByte(a) + " ); unknown code\n";
				sb.append(offset + opCode + cmd);
				
				break;
			}
			return false;
			
		default:
			return false;
		}
		return true;
		
	}

	
	private void addJumpCode(int a, int offset) {
		//System.out.println("Jump target:" + Integer.toHexString(a)  + " from opcode at offset: " + Integer.toHexString(offset));
		
		
		if (!offsetRecords.contains(a)) {
			offsetRecords.add(a);
		}
		//offsetsRecord.put(d, d);
		
	}

	private String getCommand(int x) {
		return "b["+ to2BytesStr(Integer.toHexString(x).toUpperCase())+ "];  \t\t UNPROCESSED OPCODE";
	}

	private String get2Bytes(int x) {
		String str = Integer.toHexString(x);
		while(str.length() < 4) {
			str = "0" + str;
		}
		return "0x" + str.toUpperCase();
	}
	
	private String getByte(int x) {
		String str = Integer.toHexString(x);
		while(str.length() < 2) {
			str = "0" + str;
		}
		return "0x" + str.toUpperCase();
	}
	
	private String to2BytesStr(String upperCase) {
		if(upperCase.length() == 1) {
			upperCase = "0" + upperCase;
		}
		return upperCase;
	}

	public void getMapReady() {
		offsetRecords = new ArrayList<>();
		offsetsRecord = new TreeMap<>();
		offsetTracker = new TreeMap<>();
	}
	
	public void setOffsetRecords() {
		Collections.sort(offsetRecords);
		OffsetRecords.setOffsetRecords(offsetRecords, offsetTracker);
	}
	

}
