package scriptparser;

import java.io.BufferedOutputStream;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.nio.charset.Charset;
import java.nio.file.Files;
import java.nio.file.StandardOpenOption;
import java.util.Hashtable;
import java.util.List;
import java.util.TreeMap;

import main.Menus;

public class ScriptWriter {

	BufferedOutputStream bos;
	BufferedWriter bw;
	Hashtable<String, Byte> opCodes;
	
	TreeMap<String, String> jumpTargets;
	Hashtable<String, Integer> opCodeLengths;
	
	
	ByteBuffer bb;
	
	byte cmd;

	
	public ScriptWriter(File selectFile) throws FileNotFoundException {
		loadOpCodes();
		bos = new BufferedOutputStream(new FileOutputStream(selectFile));
		
	}
	
	public ScriptWriter(File saveFile, String text) throws IOException {
		
		File file = saveFile;
		if (file.exists()) {
			file.delete();
		}
		
		bw = Files.newBufferedWriter(saveFile.toPath(), Charset.forName("UTF-8"), StandardOpenOption.CREATE_NEW, StandardOpenOption.TRUNCATE_EXISTING);
		text = text.replace("\n", System.lineSeparator());
		bw.write(text);
		
	}

	private void loadOpCodes() {
		opCodes = new Hashtable<>();
		jumpTargets = new TreeMap<>();
		
		opCodes.put("cmd_nop", (byte) 0x00);
		opCodes.put("LineMark", (byte) 0x00);
		
		opCodes.put("DisplayText", (byte) 0x01);
		opCodes.put("FlushBuffers", (byte) 0x02);
		opCodes.put("cmd_0x03", (byte) 0x03);
		opCodes.put("SetTextBox", (byte) 0x04);
		opCodes.put("SetCursorPosition", (byte) 0x05);
		opCodes.put("DisplayText_f", (byte) 0x07);
		
		opCodes.put("ExecuteScript", (byte) 0x0A);
		opCodes.put("GetSystemTime", (byte) 0x0B);
		opCodes.put("MouseOverEvent", (byte) 0x0C);
		
		opCodes.put("EnableSaveMenu", (byte) 0x0F);
		
		opCodes.put("cmd_0x13", (byte) 0x13);
		opCodes.put("GoTo", (byte) 0x14);
		opCodes.put("JumpTo", (byte) 0x15);
		opCodes.put("Return", (byte) 0x16);
		opCodes.put("cmd_0x17", (byte) 0x17);
		opCodes.put("cmd_0x18", (byte) 0x18);
		opCodes.put("ConditionalJump", (byte) 0x19);
		opCodes.put("cmd_0x1A", (byte) 0x1A);
		opCodes.put("LoadANI", (byte) 0x1E);
		opCodes.put("TweenFrames", (byte) 0x1F);
		opCodes.put("cmd_0x20", (byte) 0x20);
		opCodes.put("WaitForFrame", (byte) 0x21);
		opCodes.put("cmd_0x22", (byte) 0x22);
		opCodes.put("cmd_0x23", (byte) 0x23);
		
		opCodes.put("cmd_0x24", (byte) 0x24);
		opCodes.put("cmd_0x25", (byte) 0x25);
		
		opCodes.put("LoadANI_f", (byte) 0x27);
		opCodes.put("FadeFromBlack", (byte) 0x28);
		opCodes.put("FadeToBlack", (byte) 0x29);
		opCodes.put("FadeFromWhite", (byte) 0x2A);
		opCodes.put("FadeToWhite", (byte) 0x2B);
		opCodes.put("cmd_0x2C", (byte) 0x2C);
		opCodes.put("cmd_0x2D", (byte) 0x2D);
		opCodes.put("cmd_0x2E", (byte) 0x2E);
		opCodes.put("cmd_0x2F", (byte) 0x2F);
		opCodes.put("cmd_0x30", (byte) 0x30);
		opCodes.put("cmd_0x31", (byte) 0x31);
		opCodes.put("SetRenderSize", (byte) 0x32);
		opCodes.put("DrawImage", (byte) 0x33);
		opCodes.put("LoadCG", (byte) 0x34);
		opCodes.put("cmd_0x35", (byte) 0x35);
		opCodes.put("cmd_0x36", (byte) 0x36);
		opCodes.put("LoadCG_2", (byte) 0x37);
		
		opCodes.put("DrawImage_f", (byte) 0x39);
		
		opCodes.put("LoadAudioClip_l", (byte) 0x46);
		opCodes.put("cmd_0x47", (byte) 0x47);
		opCodes.put("PlaySFX_l", (byte) 0x48);
		opCodes.put("cmd_0x49", (byte) 0x49);
		opCodes.put("LoadBGM_2", (byte) 0x4A);
		opCodes.put("cmd_0x4B", (byte) 0x4B);
		opCodes.put("cmd_0x4C", (byte) 0x4C);
		opCodes.put("cmd_0x4D", (byte) 0x4D);
		opCodes.put("cmd_0x4E", (byte) 0x4E);
		opCodes.put("LoadAudioClip_f", (byte) 0x4F);
		opCodes.put("cmd_0x51", (byte) 0x51);  //changes after ver3
		opCodes.put("cmd_0x52", (byte) 0x52);
		
		opCodes.put("cmd_0x54", (byte) 0x54);
		opCodes.put("cmd_0x55", (byte) 0x55);
		opCodes.put("cmd_0x56", (byte) 0x56);
		opCodes.put("cmd_0x57", (byte) 0x57);
		opCodes.put("cmd_0x58", (byte) 0x58);
		opCodes.put("cmd_0x59", (byte) 0x59);
		
		opCodes.put("SetFontSize", (byte) 0x5B);
		
		opCodes.put("SetWindowSize", (byte) 0x64);
		opCodes.put("SetRenderDimensions", (byte) 0x68);
		opCodes.put("LoadSaveFile", (byte) 0x69);
		
		opCodes.put("cmd_0x6A", (byte) 0x6A);
		opCodes.put("StoreText", (byte) 0x6B);
		
		opCodes.put("TransformImageTo", (byte) 0x6E);
		opCodes.put("EnableImagePanRange", (byte) 0x6F);
		opCodes.put("WaitForTransform", (byte) 0x70);
		opCodes.put("cmd_0x71", (byte) 0x71);
		
		opCodes.put("cmd_0x73", (byte) 0x73);
		
		opCodes.put("cmd_0x78", (byte) 0x78);
		opCodes.put("LoadBGM", (byte) 0x7A);
		opCodes.put("cmd_0x7B", (byte) 0x7B);
		
		opCodes.put("cmd_0x82", (byte) 0x82);
		opCodes.put("cmd_0x83", (byte) 0x83);
		opCodes.put("cmd_0x84", (byte) 0x84);
		opCodes.put("cmd_0x8C", (byte) 0x8C);
		opCodes.put("cmd_0x8D", (byte) 0x8D);
		opCodes.put("cmd_0x96", (byte) 0x96);
		opCodes.put("cmd_0x99", (byte) 0x99);
	}

	private void calculateJumpOffsets(List<String> list) {
		loadOpCodeLengths();
		
		int positionCounter = 0;
		
		for (String s : list) {
			//skip comments and empty strings and remove extra info
			if (s.trim().isEmpty() || s.startsWith(":")) {cmd = (byte) 0x00;continue;}
			if (s.contains("|")) {s = s.substring(s.lastIndexOf("|")+1);}
			
			if (s.startsWith("@")) {
				jumpTargets.put(s.substring(1), get4HexBytes(positionCounter));
				continue;
			}
			
			positionCounter += getOpCodeLength(s);
			
			
			
		}
		
	}

	private void loadOpCodeLengths() {
		opCodeLengths = new Hashtable<>();
		
		if (Menus.getEngineVersionValue() > 2)
			opCodeLengths.put("LineMark", 3);
		opCodeLengths.put("cmd_nop", 1);
		opCodeLengths.put("DisplayText", 100);
		opCodeLengths.put("FlushBuffers", 1);
		opCodeLengths.put("cmd_0x03", 1);
		opCodeLengths.put("SetTextBox", 9);
		opCodeLengths.put("SetCursorPosition", 5);
		if (Menus.getEngineVersionValue() > 4)
			opCodeLengths.put("DisplayText_f", 109);
		opCodeLengths.put("ExecuteScript", 101);
		opCodeLengths.put("GetSystemTime", 1);
		opCodeLengths.put("MouseOverEvent", 9);
		opCodeLengths.put("EnableSaveMenu", 2);
		opCodeLengths.put("cmd_0x13", 3);
		opCodeLengths.put("GoTo", 3);
		opCodeLengths.put("JumpTo", 3);
		opCodeLengths.put("Return", 1);
		opCodeLengths.put("cmd_0x17", 3);
		opCodeLengths.put("cmd_0x18", 4);
		opCodeLengths.put("ConditionalJump", 6);
		if(Menus.getEngineVersionValue() >= 7)
			opCodeLengths.put("cmd_0x1A", 6);
		opCodeLengths.put("LoadANI", 102);
		opCodeLengths.put("TweenFrames", 6);
		opCodeLengths.put("cmd_0x20", 3);
		opCodeLengths.put("WaitForFrame", 4);
		opCodeLengths.put("cmd_0x22", 3);
		opCodeLengths.put("cmd_0x23", 3);
		opCodeLengths.put("cmd_0x24", 2);
		if (Menus.getEngineVersionValue() > 2)
			opCodeLengths.put("cmd_0x25", 1);
		if (Menus.getEngineVersionValue() > 2)
			opCodeLengths.put("LoadANI_f", 110);
		opCodeLengths.put("FadeFromBlack", 2);
		opCodeLengths.put("FadeToBlack", 2);
		opCodeLengths.put("FadeFromWhite", 2);
		opCodeLengths.put("FadeToWhite", 2);
		opCodeLengths.put("cmd_0x2C", 2);
		opCodeLengths.put("cmd_0x2D", 1);
		opCodeLengths.put("cmd_0x2E", 1);
		opCodeLengths.put("cmd_0x2F", 2);
		opCodeLengths.put("cmd_0x30", 2);
		opCodeLengths.put("cmd_0x31", 2);
		opCodeLengths.put("SetRenderSize", 9);
		opCodeLengths.put("DrawImage", 7);
		opCodeLengths.put("LoadCG", 103);
		opCodeLengths.put("cmd_0x35", 1);
		opCodeLengths.put("cmd_0x36", 1);
		opCodeLengths.put("LoadCG_2", 104);
		if (Menus.getEngineVersionValue() > 3)
			opCodeLengths.put("DrawImage_f", 8);
		opCodeLengths.put("LoadAudioClip_l", 105);
		opCodeLengths.put("cmd_0x47", 1);
		opCodeLengths.put("PlaySFX_l", 106);
		opCodeLengths.put("cmd_0x49", 1);
		opCodeLengths.put("LoadBGM_2", 107);
		opCodeLengths.put("cmd_0x4B", 1);
		opCodeLengths.put("cmd_0x4C", 2);
		opCodeLengths.put("cmd_0x4D", 2);
		opCodeLengths.put("cmd_0x4E", 2);
		if (Menus.getEngineVersionValue() > 2)
			opCodeLengths.put("LoadAudioClip_f", 111);
		
		if (Menus.getEngineVersionValue() < 3) {
			opCodeLengths.put("cmd_0x51", 5);
		}else {
			opCodeLengths.put("cmd_0x51", 6);
		}
		
		opCodeLengths.put("cmd_0x52", 1);
		opCodeLengths.put("cmd_0x54", 3);
		opCodeLengths.put("cmd_0x55", 1);
		opCodeLengths.put("cmd_0x56", 3);
		opCodeLengths.put("cmd_0x57", 1);
		opCodeLengths.put("cmd_0x58", 3);
		opCodeLengths.put("cmd_0x59", 4);
		opCodeLengths.put("SetFontSize", 4);
		opCodeLengths.put("SetWindowSize", 5);
		opCodeLengths.put("SetRenderDimensions", 5);
		opCodeLengths.put("LoadSaveFile", 3);
		if(Menus.getEngineVersionValue() >= 7)
			opCodeLengths.put("cmd_0x6A", 2);
		
		opCodeLengths.put("StoreText", 112);
		
		opCodeLengths.put("TransformImageTo", 9);
		opCodeLengths.put("EnableImagePanRange", 9);
		opCodeLengths.put("WaitForTransform",1);
		opCodeLengths.put("cmd_0x71", 1);
		opCodeLengths.put("cmd_0x73", 5);
		opCodeLengths.put("cmd_0x78", 2);
		opCodeLengths.put("LoadBGM", 108);
		opCodeLengths.put("cmd_0x7B", 1);
		
		if (Menus.getEngineVersionValue() > 2)
			opCodeLengths.put("cmd_0x82", 9);
		
		if (Menus.getEngineVersionValue() > 3)
			opCodeLengths.put("cmd_0x83", 2);
		
		if (Menus.getEngineVersionValue() > 2)
			opCodeLengths.put("cmd_0x84", 2);
		
		if (Menus.getEngineVersionValue() > 2)
			opCodeLengths.put("cmd_0x8C", 2);
		
		if (Menus.getEngineVersionValue() > 2)
			opCodeLengths.put("cmd_0x8D", 1);
		
		if (Menus.getEngineVersionValue() > 4)
			opCodeLengths.put("cmd_0x96", 5);
		
		if (Menus.getEngineVersionValue() > 5)
			opCodeLengths.put("cmd_0x99", 2);
	}

	private int getOpCodeLength(String s) {
		String temp;
		if (s.contains("Return")) {
			temp = s.substring(0, s.indexOf(";")).trim();
		} else {
			temp = s.substring(0, s.indexOf("(")).trim();
		}
		
		int x = opCodeLengths.get(temp);
		switch (x) {
		case 100: // DisplayText

			try { // had trouble here, just remembered that this can throw off offset counter
				if (s.substring(s.indexOf("\"") + 1, s.lastIndexOf("\"")).getBytes("Shift-JIS").length % 2 != 0) {
					x = s.substring(s.indexOf("\"") + 1, s.lastIndexOf("\"")).getBytes("Shift-JIS").length + 3;
				} else {
					x = s.substring(s.indexOf("\"") + 1, s.lastIndexOf("\"")).getBytes("Shift-JIS").length + 2;
				}

			} catch (UnsupportedEncodingException e) {
			}

			break;
		case 101: // ExecuteScript
			x = s.substring(s.indexOf("(") + 1, s.indexOf(".scp")).trim().length() + 2;
			break;
		case 102: // LoadANI
			x = s.substring(s.indexOf("(") + 1, s.indexOf(".ani")).trim().length() + 2;
			break;
		case 103: // LoadCG
			x = s.substring(s.indexOf("(") + 1, s.indexOf(".cg")).trim().length() + 2;
			break;
		case 104: // LoadCG_2
			x = s.substring(s.indexOf("(") + 1, s.indexOf(".cg")).trim().length() + 2;
			break;
		case 105: // LoadAudioClip_l
			x = s.substring(s.indexOf(",") + 1, s.indexOf(".pcm")).trim().length() + 3;
			break;
		case 106: // PlaySFX_l
			x = s.substring(s.indexOf(",") + 1, s.indexOf(".pcm")).trim().length() + 3;
			break;
		case 107: // LoadBGM_2
			x = s.substring(s.indexOf(",") + 1, s.indexOf(".pcm")).trim().length() + 3;
			break;
		case 108: // LoadBGM
			x = s.substring(s.indexOf(",") + 1, s.indexOf(".pcm")).trim().length() + 4;
			break;
		case 109: // DisplayText_f

			try {
				if (s.substring(s.indexOf("\"") + 1, s.lastIndexOf("\"")).getBytes("Shift-JIS").length % 2 != 0) {
					x = s.substring(s.indexOf("\"") + 1, s.lastIndexOf("\"")).getBytes("Shift-JIS").length + 3;
				} else {
					x = s.substring(s.indexOf("\"") + 1, s.lastIndexOf("\"")).getBytes("Shift-JIS").length + 2;
				}

			} catch (UnsupportedEncodingException e) {
			}
			x += 2;
			break;
		case 110:
			x = s.substring(s.lastIndexOf(",") + 1, s.indexOf(".ani")).trim().length() + 2;
			x++;
			break;
		case 111:
			x = s.substring(s.lastIndexOf(",") + 1, s.indexOf(".pcm")).trim().length() + 3;
			x++;
			break;
			
		case 112:
			
			try {
				if (s.substring(s.indexOf("\"") + 1, s.lastIndexOf("\"")).getBytes("Shift-JIS").length % 2 != 0) {
					x = s.substring(s.indexOf("\"") + 1, s.lastIndexOf("\"")).getBytes("Shift-JIS").length + 3;
				} else {
					x = s.substring(s.indexOf("\"") + 1, s.lastIndexOf("\"")).getBytes("Shift-JIS").length + 2;
				}
			} catch (UnsupportedEncodingException e) {}
			
			
			break;
			
		default:

		}
		

		//System.out.println(s + " => " + temp + "; length: " + x);
		return x;
	}

	private String get4HexBytes(int s) {
		String x = Integer.toHexString(s);
		while (x.length() < 4) {
			x = "0" + x;
		}
		return "0x" + x.toUpperCase();
	}

	public void parseScript(List<String> list ) { //Script To OpCodes
		cmd = (byte) 0xFF;
		byte[] values;
		
		AGS_Decrypt.resetDecrypt();
		
		calculateJumpOffsets(list);

		for (int x = 0; x < list.size() ;x++) {
			
			//System.out.println("working line: " + list.get(x));
			
			String temp = "";
			if (list.get(x).contains("ConditionalJump")) {
				temp = list.get(x).substring(list.get(x).lastIndexOf(",")+1, list.get(x).indexOf(")")).trim();
				list.set(x, list.get(x).replace(temp, jumpTargets.get(temp)));
			} else 
			if (list.get(x).contains("GoTo") || list.get(x).contains("JumpTo")) {
				temp = list.get(x).substring(list.get(x).lastIndexOf("(")+1, list.get(x).indexOf(")")).trim();
				list.set(x, list.get(x).replace(temp, jumpTargets.get(temp)));
			}
		}
		
		for (String s : list) {
			if (s.trim().isEmpty() || s.startsWith(":") || s.startsWith("@")) {
				cmd = (byte) 0x00;
				continue;
			}
			
			if (s.contains("|")) {
				s = s.substring(s.lastIndexOf("|")+1);
			}
			
			cmd = getOpCode(s);
			if (cmd != 0x16)
				values = getValues(s);
			else 
				values = null;
			
			try {
				// INSERT ENCRYPTION HERE
				
				if (Menus.getEngineVersionValue() >= 7) {
					cmd = AGS_Decrypt.decryptByte(cmd);
				}
				
				bos.write(cmd);
				
				if (values != null) {
					
					
					if (Menus.getEngineVersionValue() >= 7) {
						byte[] tempBytes = new byte[values.length];
						for(int i = 0; i < values.length; i++) {
							tempBytes[i] = AGS_Decrypt.decryptByte(values[i]);
						}
						values = tempBytes;
					}
					
					
					bos.write(values);
				}
				
			} catch (IOException e) {
				e.printStackTrace();
			}
		}
	}


	private byte[] getValues(String s) {
		//System.out.println("working on: " + s);
		String[] strArr;
		
		if(s.contains("DisplayText_f")) {
			strArr = new String[2];
			strArr[0] = s.substring(s.indexOf("(")+1, s.indexOf(","));
			strArr[1] = s.substring(s.indexOf("\"")+1, s.lastIndexOf("\""));
		} else if (s.contains("DisplayText") && !s.contains("DisplayText_f")){
			strArr = s.substring(s.indexOf("\"")+1, s.lastIndexOf("\"")).split("^&^&^");
		} else 
		strArr = s.substring(s.indexOf("(")+1, s.lastIndexOf(")")).split(",");
		
		
		
		
			
			
		if(s.contains("cmd_nop")) {
			return null;
		}
		
		byte[] byteArr = null;
		
		String x;
		String filename;
		String text;
		
		
		switch(cmd) {
		
		case 0x01: // Shift-JIS charSet Text
			text = strArr[0];
			try {
				if (text.getBytes("Shift-JIS").length % 2  != 0) { text += " "; }
				setUpBB(text.getBytes("Shift-JIS").length+1);
				bb.put(text.getBytes("Shift-JIS"));
				
			} catch (Exception e) {
				e.printStackTrace();
				System.exit(1);
			};
			byteArr = bb.array();
			break;
			
		case 0x07: // Shift-JIS charSet Text
			text = strArr[1];
			try {
				if (text.getBytes("Shift-JIS").length % 2  != 0) { text += " "; }
				setUpBB(text.getBytes("Shift-JIS").length+3);
				
				x = convertHexToInt(strArr[0]);
				bb.putShort((short) Integer.parseInt(x, 16));
				bb.put(text.getBytes("Shift-JIS"));
				
			} catch (Exception e) {
				e.printStackTrace();
				System.exit(1);
			};
			byteArr = bb.array();
			break;
			
		case 0x6B: // Shift-JIS charSet Text
			text = strArr[1].substring(1, strArr[1].length()-1);
			try {
				if (text.getBytes("Shift-JIS").length % 2  != 0) { text += " "; }
				setUpBB(text.getBytes("Shift-JIS").length+2);
				
				x = convertHexToInt(strArr[0]);
				bb.put((byte) Integer.parseInt(x, 16));
				bb.put(text.getBytes("Shift-JIS"));
				
			} catch (Exception e) {
				e.printStackTrace();
				System.exit(1);
			};
			byteArr = bb.array();
			break;
		
		case 0x04:  // 4 Shorts
		case 0x0C:
		case 0x32:
		case 0x6F:
		case (byte)0x82:
			setUpBB(8);
			bb.putShort((short) Integer.parseInt(strArr[0].trim()));
			bb.putShort((short) Integer.parseInt(strArr[1].trim()));
			bb.putShort((short) Integer.parseInt(strArr[2].trim()));
			bb.putShort((short) Integer.parseInt(strArr[3].trim()));
			byteArr = bb.array();
			break;
			
		case 0x05: // 2 Shorts
		case 0x64: 
		case 0x68:
			setUpBB(4);
			bb.putShort((short) Integer.parseInt(strArr[0].trim()));
			bb.putShort((short) Integer.parseInt(strArr[1].trim()));
			
			byteArr = bb.array();
			break;
			
		case 0x0A: // filename
		case 0x1E:
		case 0x34:
		case 0x37:
			filename = strArr[0].substring(0, strArr[0].lastIndexOf(".")).trim();
			setUpBB(filename.length()+1);
			bb.put(filename.getBytes());
			
			byteArr = bb.array();
			break;
			
		case 0x27: // 1 byte + filename
			filename = strArr[1].substring(0, strArr[1].lastIndexOf(".")).trim();
			setUpBB(filename.length()+2);
			
			bb.put((byte) Integer.parseInt(strArr[0].trim()));
			bb.put(filename.getBytes());
			
			byteArr = bb.array();
			break;
			
		case 0x4F: // 2 bytes + filename
			filename = strArr[2].substring(0, strArr[2].lastIndexOf(".")).trim();
			setUpBB(filename.length()+3);
			
			bb.put((byte) Integer.parseInt(strArr[0].trim()));
			bb.put((byte) Integer.parseInt(strArr[1].trim()));
			bb.put(filename.getBytes());
			
			byteArr = bb.array();
			break;
			
		case 0x13: // 1 hexShort
		case 0x14:
		case 0x15:
		case 0x54:
		case 0x56:
		case 0x58:
			setUpBB(2);
			x = convertHexToInt(strArr[0]);
			bb.putShort((short) Integer.parseInt(x, 16));
			byteArr = bb.array();
			break;	
			
			// this changes in future engine versions
		case 0x51: // 2 hexShorts
		case 0x73:
			if (Menus.getEngineVersionValue() < 3) {
				setUpBB(4);
				x = convertHexToInt(strArr[0]);
				bb.putShort((short) Integer.parseInt(x, 16));
				x = convertHexToInt(strArr[1]);
				bb.putShort((short) Integer.parseInt(x, 16));
				byteArr = bb.array();
				break;	
			} else {
				setUpBB(5);
				x = convertHexToInt(strArr[0]);
				bb.put((byte) Integer.parseInt(x, 16));
				x = convertHexToInt(strArr[1]);
				bb.putShort((short) Integer.parseInt(x, 16));
				x = convertHexToInt(strArr[2]);
				bb.putShort((short) Integer.parseInt(x, 16));
				byteArr = bb.array();
				break;	
			}
			
			
		case 0x00: // 1 Short
		case 0x20: 
		case 0x22:
			if (cmd == 0x00 && Menus.getEngineVersionValue() <= 2)
				break;
			setUpBB(2);
			bb.putShort((short) Integer.parseInt(strArr[0].trim()));
			byteArr = bb.array();
			break;	
			
		case 0x19: // 3 bytes + 1 short
			setUpBB(5);
			bb.put((byte) Integer.parseInt(strArr[0].trim()));
			bb.put((byte) Integer.parseInt(strArr[1].trim()));
			bb.put((byte) Integer.parseInt(strArr[2].trim()));
			x = convertHexToInt(strArr[3]);
			bb.putShort((short) Integer.parseInt(x, 16) );
			
			byteArr = bb.array();
			break;
			
		case 0x1F: // 1 byte + 2 shorts
			setUpBB(5);
			bb.put((byte) Integer.parseInt(strArr[0].trim()));
			bb.putShort((short) Integer.parseInt(strArr[1].trim()));
			bb.putShort((short) Integer.parseInt(strArr[2].trim()));
			byteArr = bb.array();
			break;
			
		case 0x1A:
			setUpBB(5);
			x = convertHexToInt(strArr[0]);
			bb.put((byte) Integer.parseInt(x, 16));
			x = convertHexToInt(strArr[1]);
			bb.putShort((short) Integer.parseInt(x, 16));
			x = convertHexToInt(strArr[2]);
			bb.putShort((short) Integer.parseInt(x, 16));
			
			byteArr = bb.array();
			break;
			
		case 0x21:  // 1 byte + 1 short
			setUpBB(3);
			bb.put((byte) Integer.parseInt(strArr[0].trim()));
			bb.putShort((short) Integer.parseInt(strArr[1].trim()));
			byteArr = bb.array();
			break;
			
		case 0x23: // for unknown commands:  1 hexShort
			setUpBB(2);
			x = convertHexToInt(strArr[0]);
			bb.putShort((short) Integer.parseInt(x, 16));
			
			byteArr = bb.array();
			break;
			
		case 0x0F: // 1 byte
		case 0x28:
		case 0x29:
		case 0x2A:
		case 0x2B:
		case 0x31:
		case 0x6A:
		case 0x78:
		case (byte) 0x83:
		case (byte) 0x84:
		case (byte) 0x8C:
			setUpBB(1);
			bb.put((byte) Integer.parseInt(strArr[0].trim()));
			byteArr = bb.array();
			break;
			
		case 0x24: // <--- TEST TO CONFIRM
		case 0x2C: // 1 hexByte
		case 0x2F:
		case 0x30:
		case 0x4C:
		case 0x4D:
		case 0x4E:
		case (byte) 0x99:
			setUpBB(1);
			x = convertHexToInt(strArr[0]);
			bb.put((byte) Integer.parseInt(x, 16));
			byteArr = bb.array();
			break;	
			
		case 0x33: // 3 shorts
			setUpBB(6);
			bb.putShort((short) Integer.parseInt(strArr[0].trim()));
			bb.putShort((short) Integer.parseInt(strArr[1].trim()));
			bb.putShort((short) Integer.parseInt(strArr[2].trim()));
			
			byteArr = bb.array();
			break;
			
		case 0x46: // 1 byte + 1 filename
		case 0x48:
		case 0x4A:
			filename = strArr[1].substring(0, strArr[1].lastIndexOf(".")).trim();
			setUpBB(filename.length()+2);
			
			bb.put((byte) Integer.parseInt(strArr[0].trim()));
			bb.put(filename.getBytes());
			
			byteArr = bb.array();
			break;
			
		case 0x18:// 3 bytes
		case 0x5B: 
			setUpBB(3);
			bb.put((byte) Integer.parseInt(strArr[0].trim()));
			bb.put((byte) Integer.parseInt(strArr[1].trim()));
			bb.put((byte) Integer.parseInt(strArr[2].trim()));
			byteArr = bb.array();
			break;
			
		case 0x39:// 1 hexByte + 3 hexBytes
			setUpBB(7);
			
			x = convertHexToInt(strArr[0]);
			bb.put((byte) Integer.parseInt(x, 16));
			x = convertHexToInt(strArr[1]);
			bb.putShort((short) Integer.parseInt(x, 16));
			x = convertHexToInt(strArr[2]);
			bb.putShort((short) Integer.parseInt(x, 16));
			x = convertHexToInt(strArr[3]);
			bb.putShort((short) Integer.parseInt(x, 16));
			
			byteArr = bb.array();
			break;
			
		case 0x59:// 3 hexBytes
			setUpBB(3);
			
			x = convertHexToInt(strArr[0]);
			bb.put((byte) Integer.parseInt(x, 16));
			x = convertHexToInt(strArr[1]);
			bb.put((byte) Integer.parseInt(x, 16));
			x = convertHexToInt(strArr[2]);
			bb.put((byte) Integer.parseInt(x, 16));
			byteArr = bb.array();
			break;
			
		case (byte) 0x96:// 3 hexBytes
			setUpBB(4);
			
			x = convertHexToInt(strArr[0]);
			bb.put((byte) Integer.parseInt(x, 16));
			x = convertHexToInt(strArr[1]);
			bb.put((byte) Integer.parseInt(x, 16));
			x = convertHexToInt(strArr[2]);
			bb.put((byte) Integer.parseInt(x, 16));
			x = convertHexToInt(strArr[3]);
			bb.put((byte) Integer.parseInt(x, 16));
			byteArr = bb.array();
			break;
			
		case 0x69: // 2 bytes
			setUpBB(2);
			bb.put((byte) Integer.parseInt(strArr[0].trim()));
			bb.put((byte) Integer.parseInt(strArr[1].trim()));
			byteArr = bb.array();
			break;
			
		case 0x17: // 2 hexBytes
			setUpBB(2);
			x = convertHexToInt(strArr[0]);
			bb.put((byte) Integer.parseInt(x, 16));
			x = convertHexToInt(strArr[1]);
			bb.put((byte) Integer.parseInt(x, 16));
			byteArr = bb.array();
			break;
			
		case 0x6E: //Transform unique
			setUpBB(8);
			
			String temp = strArr[0].substring(0, strArr[0].lastIndexOf("ms")).trim();
			bb.putShort((short) Integer.parseInt(temp));
			bb.putShort((short) Integer.parseInt(strArr[1].trim()));
			bb.putShort((short) Integer.parseInt(strArr[2].trim()));
			bb.putShort((short) Integer.parseInt(strArr[3].trim()));
			
			
			byteArr = bb.array();
			break;
			
		case 0x7A: // 1 short + filename
			filename = strArr[1].substring(0, strArr[1].lastIndexOf(".")).trim();
			setUpBB(filename.length()+3);
			
			bb.putShort((short) Integer.parseInt(strArr[0].trim()));
			bb.put(filename.getBytes());
			
			byteArr = bb.array();
			break;
			
		default:
			return null;
		}
		
		return byteArr;
	}

	private String convertHexToInt(String string) {
		return string.substring(string.indexOf("x")+1).trim();
	}

	private void setUpBB(int i) {
		bb = ByteBuffer.allocate(i);
		bb.position(0);
		
		for(int q = 0; q < i; q++) {
			bb.put((byte) 0);
		}
		bb.position(0);
		
		bb.order(ByteOrder.nativeOrder());
		
	}

	private byte getOpCode(String s) {
		byte opcode = (byte)0xFF;
		
		if (s.contains("(")) {
			s = s.substring(0, s.indexOf("("));
		} else if (s.contains(";")) {
			s = s.substring(0, s.indexOf(";"));
		}
		
		s = s.trim();
		
		if (opCodes.containsKey(s)) {
			opcode = opCodes.get(s);			
		}
		
		return opcode;
	}

	public void close() {
		if (bos != null) {
			try {
				bos.flush();
				bos.close();
				bos = null;
			} catch (IOException e) {
				e.printStackTrace();
			}
		} else {
			try {
				bw.flush();
				bw.close();
				bw = null;
			} catch (IOException e) {
				e.printStackTrace();
			}
		}
	}

	public void saveText(String str) {
		try {
			bos.write(str.getBytes());
			
		} catch (IOException e) {
			e.printStackTrace();
		}
	}


}
